# pylint: disable=too-few-public-methods,unused-imports,attribute-defined-outside-init,broad-exception-raise,trailing-whitespace,line-too-long,no-member,broad-exception-raised,bare-except
import json
import ast
import os
import requests
#from oauth2 import *
from xero import Xero
from xero.auth import OAuth2Credentials
from xero.exceptions import XeroForbidden
from sdc_dp_helpers.api_utilities.retry_managers import request_handler, retry_handler



def date_filter_helper(from_date: str, to_date: str, filter_field: str = None) -> str:
    """
    Custom implementation of date filters borrowed from:
    https://github.com/ClaimerApp/pyxero/blob/master/xero/basemanager.py
    """
    if not from_date:
        raise ValueError("No from_date set")

    # common date_field inside of the accounts and contacts modules is UpdatedDateUTC
    filter_field = "UpdatedDateUTC" if not filter_field else filter_field
    api_filter = filter_field + ">=DateTime(" + ",".join(from_date.split("-")) + ")"
    if to_date:
        api_filter = (
            api_filter
            + "&&"
            + filter_field
            + "<=DateTime("
            + ",".join(to_date.split("-"))
            + ")"
        )
    # end if
    return api_filter

class XeroQuotaException(Exception):
    """Class for Xero  Quota Exception"""

class Authenticator:
    """
    Xero authentication and token refresh
    """
    def __init__(self, config, creds_filepath):
        self.creds_path: str = creds_filepath
        self.config: dict = config

    def token_isvalid(self, creds: OAuth2Credentials) -> bool:
        if not isinstance(creds, OAuth2Credentials):
            raise TypeError("creds is not an object of type OAuth2Credentials")

        return creds.get_tenants()[0]["tenantId"] is not None

    def get_auth_token(self) -> OAuth2Credentials:
        """
        Consumes the client id and the previous auth processes refresh token.
        This returns an authentication token that will last 30 minutes
        to make queries the minute it is used. Or it will expire in 60 days of no use.
        The newly generated last refresh token now needs token stored for
        next use.
        PS: we receive and save the auth_token in a local dir supplied by the encapsulating project
        we never interact with s3 from inside of this class
        """
        self.client_id = self.config.get("client_id", None)
        if not self.client_id:
            raise ValueError("No client_id set")
        if (not self.client_id) or (len(self.client_id) != 32):
            raise ValueError("Invalid client_id")

        with open(self.creds_path, "r") as token_file:
            auth_token = ast.literal_eval(token_file.read())

        auth_creds = OAuth2Credentials(
            self.client_id, client_secret="", token=auth_token
        )
        return self.refresh_auth_token(auth_creds)

    def refresh_auth_token(self, auth_creds: OAuth2Credentials) -> OAuth2Credentials:
        cred = {
            "grant_type": "refresh_token",
            "refresh_token": auth_creds.token["refresh_token"],  #
            "client_id": self.config.get("client_id", None),
        }
        response = requests.post(
            "https://identity.xero.com/connect/token", cred, timeout=30
        )
        auth_token = response.json()
        # print("response:", auth_token)
        # print("auth keys", auth_token.keys())
        err_message = auth_token.get("error")
        if err_message:
            raise Exception(err_message)

        # print("Writing new Xero token to path: ", self.creds_path)
        with open(self.creds_path, "w") as outfile:
            outfile.write(json.dumps(auth_token))

        auth_creds = OAuth2Credentials(
            self.client_id, client_secret="", token=auth_token
        )
        if not self.token_isvalid(auth_creds):
            raise XeroForbidden(
                f"Error while trying to authenticate the refreshed token: {str(auth_creds)}"
            )
        return auth_creds

class XeroAPI:
    """
    Class for making Xero API calls
    """
    def __init__(self, config, creds_filepath):
        self.creds_path: str = creds_filepath
        self.config: dict = config
        self.authenticator = Authenticator(config=self.config ,creds_filepath=self.creds_path)
        #self.auth_token = self.authenticator.get_auth_token()

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 0.1)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    @retry_handler(exceptions=XeroQuotaException, total_tries=5, initial_wait=5)
    def fetch_report(self, request_params):
       
        auth_token = self.authenticator.get_auth_token()
        headers = {
            "Authorization": "Bearer " + auth_token.token["access_token"],
            "Xero-Tenant-Id": self.config["tenant_id"],
            "Accept": "application/json",
        }
        params = {
            "fromDate": self.config["start_date"],
            "toDate": self.config["end_date"],
        }
        if request_params["tracking_category"]:
            params.update(
                {
                    "trackingCategoryID": request_params["tracking_category"][
                        "TrackingCategoryID"
                    ]
                }
            )
        if request_params["TrackingOptionID"]:
            params.update({"TrackingOptionID": request_params["TrackingOptionID"]})

        if "contactId" in request_params.keys():
            params.update({"contactId": request_params["contactId"]})
        try:
            response = requests.get(
                "https://api.xero.com/api.xro/2.0/Reports/" + request_params["report_name"],
                params=params,
                headers=headers,
                timeout=30,
            )
            report = json.loads(
                    response.text.replace("\r", "").replace("\n", "").strip("'<>() "))
        except Exception as err:
            if int(response.status_code) == 429:
                os.environ['API_WAIT_TIME'] = str(10)
                raise XeroQuotaException(
                    f"Xero Quota Reached"
                    f"Status code: {err.code}, Reason: {err.reason}. "
                ) from err
            raise Exception(f"Unexpected error: {err}") from err
        return json.dumps(report)

    def format_values(self, report, option_id):
        for row in range(len(report["Reports"][0]["Rows"])):
            try:
                for row in report["Reports"][0]["Rows"][row]["Rows"]:
                    for cells in row["Cells"]:
                        cells["Attributes"][0].update({"type": row["Cells"][0]["Value"]})
            except:
                pass
        report["TrackingOptionID"] = option_id["TrackingOptionID"]
        report["TrackingOptionName"] = option_id["Name"]
        report["tenantId"] = self.config.get("tenant_id")
        return report
    def fetch_reports(self, report_name) -> dict:
        """
        Loops through reports in the config to pull each of them
        """
        data_set = {}
        if report_name not in [
            "BalanceSheet",
            "ProfitAndLoss",
            "AgedPayablesByContact",
            "AgedReceivablesByContact",
        ]:
            raise ValueError(report_name + " is not supported or does not exist.")

        auth_token = self.authenticator.get_auth_token()

        auth_token.tenant_id = self.config.get("tenant_id")

        xero_obj = Xero(auth_token)
        trackingcategories = (
            i for i in xero_obj.trackingcategories.all() if i is not None
        )
        trackingcategories_ ={}
        for tracking_category in trackingcategories:
            option_data = []
            for option_id in tracking_category["Options"]:
                request_params = {
                    "report_name": report_name,
                    "tracking_category": tracking_category,
                    "TrackingOptionID": option_id["TrackingOptionID"],
                }
                if report_name in ["BalanceSheet", "ProfitAndLoss"]:
                    file_name = (
                        report_name + "_" + tracking_category["TrackingCategoryID"]
                    )
                    result = self.fetch_report(request_params)
                    result = json.loads(result)
                    result = self.format_values(result, option_id)
                    option_data.append(result)

                elif report_name in [
                    "AgedPayablesByContact",
                    "AgedReceivablesByContact",
                ]:
                    for contact in xero_obj.contacts.filter(
                        raw="AccountNumber!=null",
                        AccountNumber__startswith="999999",
                    ):
                        request_params.update({"contactId": contact["ContactID"]})
                        file_name = (
                            report_name
                            + "_"
                            + tracking_category["TrackingCategoryID"]
                            + "_"
                            + contact["ContactID"]
                        )

                        result = self.fetch_report(request_params)
                        result = json.loads(result)
                        result = self.format_values(result, option_id)
                        option_data.append(result)
            trackingcategories_[tracking_category["Name"]]=tracking_category
            option_data.append(trackingcategories_)
            result = {
                "endpoint":report_name,
                "tenant_name":self.config["tenant_name"],
                "trackingCategoryId": tracking_category["TrackingCategoryID"],
                "data": option_data,
                "date": self.config["end_date"]}

            if option_data and option_data is not None:
                data_set.update({file_name: result})
        return data_set
    
    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 0.1)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
       backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    def run_request(self, xero_client, api_object, request):
        """
        Run the API request that consumes a request payload and site url.
        This separates the request with the request handler from the rest of the logic.
        """
        # To Do Handle API Errors
        api_call = getattr(xero_client, api_object)
        # XeroRateLimitExceeded
        return api_call.filter(
            raw=date_filter_helper(request["from_date"], request["to_date"]),
            page=request["page"],
        )

    def fetch_modules(self,module)-> dict:
        """
        Consumes a .yaml config file and loops through the date and url
        to return relevant data from Xero API.
        """
        auth_token = self.authenticator.get_auth_token()
        auth_token.tenant_id =  self.config["tenant_id"]
        xero = Xero(auth_token)
        data_set = {}
        if module not in [
            "contacts",
            "accounts",
            "invoices",
            "banktransactions",
            "manualjournals",
            "purchaseorders",
        ]:
            raise ValueError(module + " is not supported or does not exist.")
        data_set[module] = []
        prev_response = None
        page = 1
        while True:
            response = self.run_request(
                xero_client=xero,
                api_object=module,
                request={
                    "from_date": self.config.get("from_date"),
                    "to_date": self.config.get("to_date"),
                    "page": page,
                },
            )
            if len(response) < 1:
                print("Request returned empty payload. breaking...")
                break
            if response == prev_response:
                print("Request returned copy of last payload. breaking...")
            data_set[module] += [
                json.loads(
                    json.dumps(response_obj, indent=4, sort_keys=True, default=str)
                )
                for response_obj in response
            ]
            # ensure the token is still fresh
            auth_token = self.get_auth_token()
            prev_response = response
            page += 1
        return data_set



"""
    CUSTOM FALCON READER CLASSES
"""
# pylint: disable=too-few-public-methods,import-error,unused-import,too-many-locals,arguments-differ

from typing import Generator, List, Dict
from datetime import datetime

import requests

from sdc_dp_helpers.base_readers import BaseReader
from sdc_dp_helpers.api_utilities.date_managers import date_range_iterator
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.falcon.falcon_sdk import FalconAPICall, FalconQuotaException


class FalconReader(BaseReader):
    """Falcon Reader class

    Args:
        BaseReader (_type_): API Reader Class
    """

    def __init__(self, creds_file: str, config_file=None):
        """creds_file has crecdentials config_file has api pull configs"""
        self._creds: dict = load_file(creds_file, "yml")
        self._config: dict = load_file(config_file, "yml")
        self.session: requests.Session = requests.Session()
        self.success = []
        self.service: FalconAPICall = self._get_auth()
        self.start = 0

    def get_channel_ids(self):
        return self.service.get_id_lookup()

    def _get_auth(self):
        self.service = FalconAPICall(
            session=self.session, creds=self._creds, config=self._config
        )

        return self.service

    def _query_handler(
        self, start_date: str, end_date: str, network: str, channel_ids: dict
    ) -> List[Dict]:

        data = self.service.get_data(
            start_date=start_date,
            end_date=end_date,
            network=network,
            channel_ids=channel_ids,
        )
        return data

    def run_query(self) -> Generator:
        """Calls the Query Handler"""
        endpoint_name = self._config.get("endpoint_name")
        channel_ids = self.service.get_channel_ids()
        network = self._config["networks"]

        date_iterator = date_range_iterator(
            start_date=self._config["since"],
            end_date=self._config["until"],
            interval="1_day",
            end_inclusive=False,
            time_format="%Y-%m-%dT00:00:00Z",
        )
        try:
            for start_date, end_date in date_iterator:
                payload: List = []
                payload = self._query_handler(
                    start_date=start_date,
                    end_date=end_date,
                    network=network,
                    channel_ids=channel_ids,
                )
                date = datetime.strftime(datetime.strptime(
                    start_date, "%Y-%m-%dT%H:%M:%SZ"), "%Y-%m-%d")
                if payload:
                    yield {
                        "networks": network,
                        "date": date,
                        "data": payload,
                    }
                    self.start = 0
                    self.is_success()
                else:
                    self.not_success()
                    print(
                        f"No data for endpoint {endpoint_name} for date : {date}")
        except FalconQuotaException as exp:
            raise exp



"""
    Retry Managers:
        - retry decorators
        - failure mode handlers
"""
# pylint: disable=inconsistent-return-statements, disable=line-too-long
import os
import time
from functools import wraps
from random import random

from typing import Union
from sdc_helpers.slack_helper import SlackHelper


def retry_handler(
    exceptions,
    total_tries: int = 4,
    initial_wait: Union[int, float] = 0.5,
    backoff_factor: Union[int, float] = 2,
    should_raise: bool = False,
):
    """
    Decorator - managing API failures

    args:
        exceptions (Exception): Exception instance or list of
        Exception instances to catch & retry
        total_tries (int): Total retry attempts
        initial_wait (float): initial delay between retry attempts in seconds
        backoff_factor (int): multiplier used to further randomize back off
        should_raise (bool): Raise exception after all retires fail

    return:
        wrapped function's response
    """

    def retry_decorator(function):
        @wraps(function)
        def func_with_retries(*args, **kwargs):
            """wrapper function to decorate function with retry functionality"""
            _tries, _delay = total_tries, initial_wait
            while _tries > 0:
                try:
                    print(f"Attempt {total_tries + 1 - _tries}")
                    return function(*args, **kwargs)
                except exceptions as exception:
                    # get logger message
                    _delay = float(os.environ.get('API_WAIT_TIME', _delay))
                    if _tries == 1:
                        slack_helper = SlackHelper()
                        slack_helper.send_critical(
                            message=str(
                                f"Function: {function.__name__}\n"
                                f"Failed despite best efforts after {total_tries} tries\n"
                                f"Exception {exception}."
                            )
                        )

                        if should_raise:
                            raise exception
                    else:
                        msg = str(
                            f"Function: {function.__name__}\n"
                            f"Exception {exception}.\n"
                            f"Retrying in {_delay} seconds!"
                        )
                        # log with print
                        print(msg)
                    # decrement _tries
                    _tries -= 1
                    # pause
                    time.sleep(_delay)
                    # increase delay by backoff factor
                    _delay = _delay * backoff_factor

        return func_with_retries

    return retry_decorator


def request_handler(
    wait: Union[int, float] = 0, backoff_factor: Union[int, float] = 0.01, backoff_method: str = "random"
):
    """
    Decorator - managing API failures

    Args:
        wait (float, optional): [description].
            Defaults to 100 seconds for GoogleAnalytics.
            Initial delay between retry attempts in seconds
        backoff_factor (float, optional):
            [description]. Defaults to 0.1.
            Multiplier used to further randomize delay
        backoff_method (str, optional):
            [description]. Defaults to 'random'. hHow to apply backoff
    return:
        wrapped function's response
    """

    def retry_decorator(function):
        @wraps(function)
        def func_with_retries(*args, **kwargs):
            """wrapper function to decorate function with retry functionality"""
            if backoff_method == "random":
                _delay = wait * (1 + backoff_factor * random())
            else:
                _delay = wait

            print(f"Waiting {_delay} seconds before attempt")
            time.sleep(_delay)
            return function(*args, **kwargs)

        return func_with_retries

    return retry_decorator
